#!/usr/bin/env python

"""
OpenID initiation resources which redirect clients to an OpenID provider.

Copyright (C) 2004, 2005, 2006, 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

import WebStack.Generic
import libxml2dom
import cgi # for escape

class OpenIDInitiationUtils:

    "Utilities for OpenID initiation screens which may be inherited."

    openid_ns = "http://specs.openid.net/auth/2.0"

    def __init__(self, openid_mode=None, use_redirect=1, urlencoding=None):

        """
        Initialise the resource.

        The optional 'openid_mode' parameter may be set to "checkid_immediate"
        or "checkid_setup" (the default).

        If the optional 'use_redirect' flag is set to a false value (which is
        not the default), a confirmation screen is given instead of immediately
        redirecting the user to the OpenID provider.

        The optional 'urlencoding' parameter allows a special encoding to be
        used in producing the redirection path.
        """

        self.openid_mode = openid_mode or "checkid_setup"
        self.use_redirect = use_redirect
        self.urlencoding = urlencoding

    def get_redirect_url(self, trans, app, claimed_identifier, provider, local_identifier):

        # NOTE: Should consider the special "select" mode for identity.

        return "%s?openid.ns=%s&openid.mode=%s&openid.return_to=%s&openid.claimed_id=%s&openid.identity=%s" % (
            trans.encode_url_without_query(provider, self.urlencoding),
            trans.encode_path(self.openid_ns, self.urlencoding),
            trans.encode_path(self.openid_mode, self.urlencoding),
            trans.encode_path(app, self.urlencoding),
            trans.encode_path(claimed_identifier, self.urlencoding),
            trans.encode_path(local_identifier, self.urlencoding)
            )

    def get_provider_url(self, trans, identity):

        """
        Return the claimed identifier, provider URL and local identifier for the
        authenticating user using the given 'trans' and 'identity'.

        See:
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.7.3
        """

        if identity.startswith("xri://"):
            identity = openid[6:]

        # NOTE: Not yet discovering XRI providers.

        if identity[0] in ("=", "@", "+", "$", "!", "("):
            pass
        else:
            if not identity.startswith("http"):
                identity = "http://" + identity

            # Obtain a provider url from a resource at the stated URL.

            doc = libxml2dom.parseURI(trans.encode_url_without_query(identity), html=1)
            provider_links = doc.xpath("/html/head/link[contains(@rel, 'openid2.provider')]/@href")
            local_ids = doc.xpath("/html/head/link[contains(@rel, 'openid2.local_id')]/@href")
            if provider_links:
                if local_ids:
                    return identity, provider_links[0].nodeValue, local_ids[0].nodeValue
                else:
                    return identity, provider_links[0].nodeValue, None

        return identity, None, None

class OpenIDInitiationResource(OpenIDInitiationUtils):

    "A resource providing an OpenID initiation screen."

    def __init__(self, openid_mode=None, use_redirect=1, urlencoding=None, encoding=None):

        """
        Initialise the resource.

        The optional 'openid_mode' parameter may be set to "checkid_immediate"
        or "checkid_setup" (the default).

        If the optional 'use_redirect' flag is set to a false value (which is
        not the default), a confirmation screen is given instead of immediately
        redirecting the user to the OpenID provider.

        The optional 'urlencoding' parameter allows a special encoding to be
        used in producing the redirection path.

        The optional 'encoding' parameter allows a special encoding to be used
        in producing the initiation pages.

        To change the pages employed by this resource, either redefine the
        'initiation_page' and 'success_page' attributes in instances of this class or
        a subclass, or override the 'show_initiation' and 'show_success' methods.
        """

        OpenIDInitiationUtils.__init__(self, openid_mode, use_redirect, urlencoding)
        self.encoding = encoding

    def respond(self, trans):

        "Respond using the transaction 'trans'."

        app = get_target(trans, self.urlencoding, self.encoding)

        # Check for a submitted initiation form.

        fields_body = trans.get_fields_from_body(self.encoding)

        if fields_body.has_key("initiate") and fields_body.has_key("identity"):
            self.check_identity(trans, fields_body, app)
            # The above method does not return.

        # Otherwise, show the initiation form.

        self.show_initiation(trans, app)

    def check_identity(self, trans, fields, app):

        """
        Check the identity found through 'trans' and 'fields', using 'app' and
        discovered information about the identity to redirect to the provider.
        """

        claimed_identifier, provider, local_identifier = self.get_provider_url(trans, fields["identity"][0])
        if provider is not None:
            self.redirect_to_provider(trans, app, claimed_identifier, provider, local_identifier)

    def redirect_to_provider(self, trans, app, claimed_identifier, provider, local_identifier):

        """
        Redirect the client using 'trans' and the given 'app',
        'claimed_identifier', 'provider' and 'local_identifier' details.

        See:
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.5.2
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.9
        """

        url = self.get_redirect_url(trans, app, claimed_identifier, provider, local_identifier)

        # Show the success page anyway.
        # Offer a POST-based form for redirection.

        self.show_success(trans, provider, app, claimed_identifier, local_identifier)

        # Redirect to the OpenID provider URL.

        if self.use_redirect:
            trans.redirect(url)
        else:
            raise WebStack.Generic.EndOfResponse

    def show_initiation(self, trans, app):

        """
        Writes a initiation screen using the transaction 'trans', including details
        of the 'app' which the client was attempting to access.
        """

        trans.set_content_type(WebStack.Generic.ContentType("text/html", self.encoding or trans.default_charset))
        out = trans.get_response_stream()
        out.write(self.initiation_page % cgi.escape(app))

    def show_success(self, trans, provider, app, claimed_identifier, local_identifier):

        """
        Writes a success screen using the transaction 'trans', including details
        of the OpenID 'provider', the 'app' URL, 'claimed_identifier' and
        'local_identifier'.
        """

        trans.set_content_type(WebStack.Generic.ContentType("text/html", self.encoding or trans.default_charset))
        out = trans.get_response_stream()
        out.write(self.success_page % tuple(map(cgi.escape, (
            provider, self.openid_ns, self.openid_mode, app, claimed_identifier, local_identifier)
            )))

    initiation_page = """
<html>
  <head>
    <title>Authenticate via OpenID</title>
  </head>
  <body>
    <h1>Authenticate via OpenID</h1>
    <form method="POST" name="openid_identifier">
      <p>OpenID Identifier (URL): <input name="identity" type="text" size="32"/></p>
      <p><input name="initiate" type="submit" value="Login" /></p>
      <input name="app" type="hidden" value="%s" />
    </form>
  </body>
</html>
"""

    success_page = """
<html>
  <head>
    <title>Authenticate via OpenID</title>
  </head>
  <body>
    <h1>Authenticate via OpenID</h1>
    <form action="%s" method="POST" name="openid_redirect">
      <input name="openid.ns" type="hidden" value="%s" />
      <input name="openid.mode" type="hidden" value="%s" />
      <input name="openid.return_to" type="hidden" value="%s" />
      <input name="openid.claimed_id" type="hidden" value="%s" />
      <input name="openid.identity" type="hidden" value="%s" />
      <p>Please proceed to the OpenID provider: <input name="proceed" type="submit" value="Proceed!" /></p>
    </form>
  </body>
</html>
"""

# General functions.

def get_target(trans, urlencoding=None, encoding=None):

    """
    Return the application for 'trans' using the optional 'urlencoding' (or path
    encoding) and request body 'encoding'.
    """

    fields_path = trans.get_fields_from_path(urlencoding)
    fields_body = trans.get_fields_from_body(encoding)

    # NOTE: Handle missing redirects better.

    if fields_body.has_key("app"):
        apps = fields_body["app"]
        app = apps[0]
    elif fields_path.has_key("app"):
        apps = fields_path["app"]
        app = apps[0]
    else:
        app = u""

    return app

# vim: tabstop=4 expandtab shiftwidth=4
